"""The ncs-uml creates plantUML files for the given YANG file and its dependencies"""

__name__ = 'ncs-uml'
__version__ = '1.2.2'
__date__ = '2023-06-23'
__description__ = "The ncs-uml creates plantUML files for the given YANG file and its dependencies"